# This file is part of reLyX, copyright (c) 1998 Amir Karger
# You are free to use and modify this code under the terms of
# the GNU General Public Licence version 2 or later.

package BasicLyX;
# This package includes subroutines to translate "clean" LaTeX to LyX.
# It translates only "basic" stuff, which means it doesn't do class-specific
#     things. (It uses the TeX parser Text::TeX)
use strict;

use RelyxTable; # Handle LaTeX tables
#################### PACKAGE-WIDE VARIABLES ###########################
my $debug_on; # is debugging on?

######
#Next text starts a new paragraph?
# $INP = 0 for plain text not starting a new paragraph
# $INP = 1 for text starting a new paragraph, so that we write a new
#    \layout command and renew any font changes for the new paragraph
# Starts as 1 cuz if the first text in the document is plain text
#    (not, e.g., in a \title command) it will start a new paragraph
my $IsNewParagraph = 1;
my $OldINP; #Save $IsNewParagraph during footnotes

# HACK to protect spaces within optional argument to \item
my $protect_spaces = 0;

# $MBD = 1 if we're in a list, but haven't seen an '\item' command
#    In that case, we may need to start a nested "Standard" paragraph
my $MayBeDeeper = 0;
my $OldMBD; #Save $MBD during footnotes -- this should very rarely be necessary!

# Stack to take care of environments like Enumerate, Quote
# We need a separate stack for footnotes, which have separate layouts etc.
# Therefore we use a reference to an array, not just an array
my @LayoutStack = ("Standard"); #default if everything else pops off
my $CurrentLayoutStack = \@LayoutStack;

# Status of various font commands
# Every font characteristic (family, series, etc.) needs a stack, because
#    there may be nested font commands, like \textsf{blah \texttt{blah} blah}
# CurrentFontStatus usually points to the main %FontStatus hash, but
#     when we're in a footnote, it will point to a temporary hash
my %FontStatus = (
    '\emph' => ["default"],
    '\family' => ["default"],
    '\series' => ["default"],
    '\shape' => ["default"],
    '\bar' => ["default"],
    '\size' => ["default"],
);
my $CurrentFontStatus = \%FontStatus;

# Currently aligning paragraphs right, left, or center?
my $CurrentAlignment = "";
my $OldAlignment; # Save $AS during footnotes

# Global variables for copying tex stuff
my $in_tex_mode = 0; # Copying known but untranslatable token in tex mode?
my $tex_mode_string; # string we accumulate tex mode stuff in
my $tex_start; # token that causes reLyX to start copying in tex mode

#############  INFORMATION ABOUT LATEX AND LYX   #############################
# LyX translations of LaTeX font commands
my %FontTransTable = (
   # Font commands
   '\emph' => "\n\\emph on\n",
   '\underline' => "\n\\bar under\n",
   '\textbf' => "\n\\series bold\n",
   '\textmd' => "\n\\series medium\n",
   '\textsf' => "\n\\family sans\n",
   '\texttt' => "\n\\family typewriter\n",
   '\textrm' => "\n\\family roman\n",
   '\textsc' => "\n\\shape smallcaps\n",
   '\textsl' => "\n\\shape slanted\n",
   '\textit' => "\n\\shape italic\n",
   '\textup' => "\n\\shape up\n",

# Font size commands
   '\tiny' => "\n\\size tiny\n",
   '\scriptsize' => "\n\\size scriptsize\n",
   '\footnotesize' => "\n\\size footnotesize\n",
   '\small' => "\n\\size small\n",
   '\normalsize' => "\n\\size default\n",
   '\large' => "\n\\size large\n",
   '\Large' => "\n\\size Large\n",
   '\LARGE' => "\n\\size LARGE\n",
   '\huge' => "\n\\size huge\n",
   '\Huge' => "\n\\size Huge\n",
);

# Things LyX implements as "Floats"
my %FloatTransTable = (
   # Footnote, Margin note
   '\footnote' => "\n\\begin_float footnote\n",
   '\thanks' => "\n\\begin_float footnote\n", # thanks is same as footnote
   '\marginpar' => "\n\\begin_float margin\n",
);
# Environments that LyX implements as "floats"
my %FloatEnvTransTable = (
    "table" => "\n\\begin_float tab\n",
    "table*" => "\n\\begin_float wide-tab\n",
);

# Simple LaTeX tokens which are turned into small pieces of LyX text
# Since these tokens can appear in the midst of text, we may have to, e.g.,
#     start a new paragraph right before one of them.
my %TextTokenTransTable = (
    # LaTeX escaped characters
    '\_' => '_',
    '\%' => '%',
    '\$' => '$',
    '\&' => '&',
    '\{' => '{',
    '\}' => '}',
    '\#' => '#',
    # \i and \j are used in accents. LyX doesn't recognize them in plain
    #    text. Hopefully, this isn't a problem.
    '\i' => '\i',
    '\j' => '\j',

    # Misc simple LaTeX tokens
    '~' => "\n\\protected_separator \n",
    '@' => "@", # TeX.pm considers this a token, but it's not really
    '\@' => "\\SpecialChar \\@",
    '\ldots' => "\n\\SpecialChar \\ldots\{\}\n",
    '\-' => "\\SpecialChar \\-\n",
    '\LaTeX' => "LaTeX",
    '\TeX' => "TeX",
    '\LyX' => "LyX",
    '\noindent' => "\n\\noindent\n",
    '\tableofcontents' =>
             "\n\\begin_inset LatexCommand \\tableofcontents\n\n\\end_inset\n",
);

# LyX translations of some plain LaTeX text (TeX parser won't recognize it
#     as a Token, so we need to translate the Text::TeX::Text token.)
my %TextTransTable = (
    # Double quotes
    "``" => "\n\\begin_inset Quotes eld\n\\end_inset\n\n",
    "''" => "\n\\begin_inset Quotes erd\n\\end_inset\n",

    # Tokens that don't start with a backslash, so parser won't recognize them
    # (LyX doesn't support them, so we just print them in TeX mode)
    '?`' => "\n\\latex latex\n?`\n\\latex default\n",
    '!`' => "\n\\latex latex\n!`\n\\latex default\n",
);

# Things that LyX translates as "LatexCommand"s
my $LatexCommands = "\\\\(ref|pageref|label|cite)";

# Accents. Most of these take an argument -- the thing to accent
# (\l and \L are handled as InsetLatexAccents, so they go here too)
my $AccentTokens = "\\\\[`'^#~=.uvHtcdblL\"]";

# Environments which describe justifying (converted to LyX \align commands)
#    and the corresponding LyX commands
my %AlignEnvironments = (
    "center" => "\n\\align center",
    "flushright" => "\n\\align right",
    "flushleft" => "\n\\align left",
);

# Math environments are copied verbatim
my $MathEnvironments = "(math|equation|displaymath|eqnarray(\\*)?)";
# ListLayouts may have standard paragraphs nested inside them.
my $ListLayouts = "Itemize|Enumerate|Description";

#####################   PARSER INVOCATION   ##################################
sub call_parser {
# This subroutine calls the TeX parser & translator
# Before it does that, it does lots of setup work to get ready for parsing.
# Arg0 is the file to read (clean) LaTeX from
# Arg1 is the file to write LyX to
# Arg2 is the file to read layouts from (e.g., in LYX_DIR/layouts/foo.layout)

    my ($InFileName, $OutFileName) = (shift,shift);

# Before anything else, set the package-wide variables based on the
#    user-given flags
    # opt_d is set to 1 if '-d' option given, else (probably) undefined
    $debug_on = (defined($main::opt_d) && $main::opt_d);

    # Hash of tokens passed to the TeX parser
    # Many values are $Text::TeX::Tokens{'\operatorname'}, which has
    #    Type=>report_args and count=>1
    # Note that we don't have to bother putting in tokens which will be simply
    #    translated (e.g., from %TextTokenTransTable).
    my %MyTokens = ( 
	'{' => $Text::TeX::Tokens{'{'},
	'}' => $Text::TeX::Tokens{'}'},
	'\begin' => $Text::TeX::Tokens{'\begin'},
	'\end' => $Text::TeX::Tokens{'\end'},

	# Lots of other commands will be made by ReadCommands:Merge
	# by reading the LaTeX syntax file

	# Font sizing commands (local)
	'\tiny' => {Type => 'local'},
	'\small' => {Type => 'local'},
	'\scriptsize' => {Type => 'local'},
	'\footnotesize' => {Type => 'local'},
	'\small' => {Type => 'local'},
	'\normalsize' => {Type => 'local'},
	'\large' => {Type => 'local'},
	'\Large' => {Type => 'local'},
	'\LARGE' => {Type => 'local'},
	'\huge' => {Type => 'local'},
	'\Huge' => {Type => 'local'},

	# Tokens to ignore (which make a new paragraph)
	# Just pretend they actually ARE new paragraph markers!
	'\maketitle' => {'class' => 'Text::TeX::Paragraph'},
    );
    
    # Now add to MyTokens all of the commands that were read from the
    #    commands file by package ReadCommands
    &ReadCommands::Merge(\%MyTokens);

# Here's the actual subroutine. The above is all preparation
    # Output LyX file
    warn "Basic LaTeX to LyX translation... \n";
    open (OUTFILE,">$OutFileName");

    # Open the file to turn into LyX.
    my $infile = new Text::TeX::OpenFile $InFileName,
	'defaultact' => \&basic_lyx,
	'tokens' => \%MyTokens;

    # Process what's left of the file (everything from \begin{document})
    $infile->process;

    # Last line of the LyX file
    print OUTFILE "\n\\the_end\n";
    close OUTFILE;
    warn "Done with basic translation\n";
    return;
} # end subroutine call_parser

##########################   MAIN TRANSLATOR SUBROUTINE   #####################
sub basic_lyx {
# This subroutine is called by Text::TeX::process each time subroutine
#     eat returns a value.
# Argument 0 is the return value from subroutine eat
# Argument 1 is the Text::TeX::OpenFile (namely, $TeXfile)
    my $eaten = shift;
    my $fileobject = shift;

    # This handles most but maybe not all comments
    # THere shouldn't be any if we've run CleanTeX.pl
    if (defined $eaten->[1]) {print "Comment: $eaten->[1]" if $debug_on};

    my $type = ref($eaten);
    print "$type " if $debug_on;

    # This loop is basically just a switch. However, making it a for
    #    (1) makes $_ = $type (convenient)
    #    (2) allows us to use things like next and last
    TYPESW: for ($type) {

        # some pre-case work
        s/^Text::TeX:://o or die "unknown token?!";
	my ($outstr, $env, $layout, $command, $dummy, $tok, @arr);
	my ($thistable);
        
	# If we're copying a recognized but untranslatable token in tex mode...
	if ($in_tex_mode) {
	    # TT::Paragraph can't be printed, but we don't really want to
	    # set $INP in the middle of tex mode, so just ignore it
	    &copy_latex_known($eaten,$fileobject) unless /^Paragraph$/;
	    #don't do the whole TYPESW
	    last TYPESW;
	}

        # Handle blank lines.
        if (m/^Paragraph$/o) {
	    # $INP <>0 means We will need a new layout command
	    $IsNewParagraph = 1;

	    # $MBD means start a begin_deeper within list environments
	    #    unless there's an \item command
	    $MayBeDeeper = 1;

            last TYPESW;
        }

	# This is usually the text that makes up (a part of) the token
        $outstr = $eaten->[0];

	# If, e.g., there's just a comment in this token, don't do anything
	# This actually shouldn't happen if CleanTeX has already removed them
	last TYPESW if !defined $outstr;
        
        # Handle LaTeX tokens
        if (/^Token$/o) {

	    print "'$outstr' " if $debug_on;

	    # Tokens which turn into a bit of LyX text
	    if (exists $TextTokenTransTable{$outstr}) {
	        &CheckForNewParagraph; #Start new paragraph if necessary

		# Print the translated text
		print OUTFILE "$TextTokenTransTable{$outstr}";

	    # Math -- copy verbatim until you're done
	    } elsif ($outstr eq '\(' || $outstr eq '\[') {
		print "\nCopying math beginning with '$outstr'\n" if $debug_on;
		# copy everything until end text
		$dummy = &copy_verbatim($fileobject, $eaten);

		&CheckForNewParagraph; # math could be first thing in a par
		print OUTFILE "\n\\begin_inset Formula $outstr ";
		print $dummy if $debug_on;
		print OUTFILE $dummy;

	    } elsif ($outstr eq '\)' || $outstr eq '\]') {
	        # end math
		print OUTFILE "$outstr\n\\end_inset\n";
		print "\nDone copying math ending with '$outstr'" if $debug_on;

	    # Items in list environments
	    } elsif ($outstr eq '\item') {
		
		# What if we had a nested "Standard" paragraph?
		# Then write \end_deeper to finish the standard layout
		#     before we write the new \layout ListEnv command
		if ($$CurrentLayoutStack[-1] eq "Standard") {
		    pop (@$CurrentLayoutStack); # take "Standard" off the stack
		    print OUTFILE "\n\\end_deeper\n";
		    print "\nCurrent Layout Stack: @$CurrentLayoutStack"
		          if $debug_on;
		} # end deeper if

		# Upcoming text (the item) will be a new paragraph, 
		#    requiring a new layout command based on whichever
		#    kind of list environment we're in
		$IsNewParagraph = 1;

		# But since we had an \item command, DON'T nest a
		#    deeper "Standard" paragraph in the list
		$MayBeDeeper = 0;

		# Check for an optional argument to \item
		# If so, within the [] we need to protect spaces
		if (($dummy = $fileobject->lookAheadToken) &&
		    ($dummy =~ /\s*\[/)) {
		    $fileobject->eatFixedString('\['); # eat the [
		    $protect_spaces = 1;
		}

	    # Font sizing commands
	    # (Other font commands are TT::BegArgsTokens because they take
	    #     arguments. Font sizing commands are 'local' TT::Tokens)
	    } elsif (exists $FontTransTable{$outstr}) {
		$command = $FontTransTable{$outstr}; #e.g., '\size large'

		if (! $IsNewParagraph) {
		    print OUTFILE "$command";
		} #otherwise, wait until we've printed the new paragraph command

		# Store the current font change
		($dummy = $command) =~ s/\s*(\S+)\s+(\w+)\s*/$1/;
		die "Font command error" if !exists $$CurrentFontStatus{$dummy};
		push (@{$CurrentFontStatus->{$dummy}}, $2);
		print "\nCurrent $dummy Stack: @{$CurrentFontStatus->{$dummy}}"
		      if $debug_on;

	    # Table stuff
	    } elsif ($outstr eq '&') {
		if ($thistable = &RelyxTable::in_table) {
		    print OUTFILE "\n\\newline\n";
		    $thistable->nextcol;
		} else {warn "& is illegal outside a table!"}

	    } elsif ($outstr eq '\\\\' || $outstr eq '\\newline') {
		&CheckForNewParagraph; # could be at beginning of par?
                print OUTFILE "\n\\newline\n";

		# If we're in a table, \\ means add a row to the table
		if ($thistable = &RelyxTable::in_table) {
		    $thistable->addrow;
		}

	    } elsif ($outstr eq '\hline') {
		if ($thistable = &RelyxTable::in_table) {
		    # hcline does hline if no arg is given
		    $thistable->hcline;
		} else {warn "\\hline is illegal outside a table!"}

	    } elsif ($outstr eq '\cline') {
		if ($thistable = &RelyxTable::in_table) {
		    # read the argument to \cline
		    $tok = $fileobject->eatBalanced; # eat w/out braces
		    $dummy = $tok->[0]->print; #Group should have just one token
		    # hcline does cline if an arg is given
		    $thistable->hcline($dummy);
		} else {warn "\\cline is illegal outside a table!"}

	    } elsif ($outstr eq '\multicolumn') {
		if ($thistable = &RelyxTable::in_table) {
		    # Eat the first two arguments to \multicolumn command
		    # The first has to be simple, so we use eatBalanced
		    # The second could have @{}'s in it, so we use eatGroup
		    $dummy = $fileobject->eatBalanced;
		    $dummy = $dummy->[0]->print;
		    $tok = $fileobject->eatGroup;
		    $thistable->multicolumn($dummy, $tok);
		} else {warn "\\multicolumn is illegal outside a table!"}

	    # Otherwise it's an unknown token, which must be copied
	    #     in TeX mode, along with its arguments, if any
	    } else {
		if (defined($eaten->relyx_args($fileobject))) {
		   &copy_latex_known($eaten, $fileobject);
		} else { # it's not in MyTokens
		    &copy_latex_unknown($eaten, $fileobject);
		}
	    }

            last TYPESW;
        }
        
	# Handle tokens that take arguments, like \section{},\section*{}
	if (/^BegArgsToken$/) {
	    $env = $eaten->[0]->[0]->[0];
	    print "$env" if $debug_on;

	    # Handle things that LyX translates as a "LatexCommand" inset
	    # \ref{blah} -> '\begin_inset LatexCommand \ref{blah}\n\n\end_inset'
	    if ($env =~ /^$LatexCommands$/o) {
		&CheckForNewParagraph; #Start new paragraph if necessary

	        print OUTFILE "\n\\begin_inset LatexCommand $env\{";
	        last TYPESW; # skip to the end of the switch
	    }

	    # Sectioning and Title environments (using a LyX \layout command)
	    if (exists $ReadCommands::ToLayout->{$env}) {
		&ConvertToLayout($env);
		last TYPESW; #done translating

	    # Font characteristics
	    } elsif (exists $FontTransTable{$env}) {
		my $dum2;
	        $command = $FontTransTable{$env};
	        ($dummy, $dum2) = ($command =~ /(\S+)\s+(\w+)/);

		# HACK so that "\emph{hi \emph{bye}}" yields unemph'ed "bye"
		if ( ($dummy eq "\\emph") && 
		     ($CurrentFontStatus->{$dummy}->[-1] eq "on")) {
		       $dum2 = "default"; # "on" instead of default?
		       $command =~ s/on/default/;
		}

		# If we're about to start a new paragraph, save writing
		#    this command until *after* we write '\layout Standard'
		if (! $IsNewParagraph) {
		    print OUTFILE "$command";
		}

		# Store the current font change
		die "Font command error" if !exists $$CurrentFontStatus{$dummy};
		push (@{$CurrentFontStatus->{$dummy}}, $dum2);


	    # Handle footnotes and margin notes
	    # Make a new font table & layout stack which will be local to the 
	    #    footnote or marginpar
	    } elsif (exists $FloatTransTable{$env}) {
	        $command = $FloatTransTable{$env};

		# Open the footnote
		print OUTFILE "$command";

		# Make $CurrentFontStatus point to a new (anonymous) font table
	        $CurrentFontStatus =  {
		    '\emph' => ["default"],
		    '\family' => ["default"],
		    '\series' => ["default"],
		    '\shape' => ["default"],
		    '\bar' => ["default"],
		};

		# And make $CurrentLayoutStack point to a new (anon.) stack
		$CurrentLayoutStack = ["Standard"];

		# Store whether we're at the end of a paragraph or not
		#    for when we get to end of footnote AND 
		# Note that the footnote text will be starting a new paragraph
		# Also store the current alignment (justification)
		$OldINP = $IsNewParagraph; $OldMBD = $MayBeDeeper;
		$OldAlignment = $CurrentAlignment;
		$IsNewParagraph = 1;
		$MayBeDeeper = 0; #can't be deeper at beginning of footnote
		$CurrentAlignment = "";

	    # Accents
	    } elsif ($env =~ m/^$AccentTokens$/) {
		&CheckForNewParagraph; # may be at the beginning of a par

		print OUTFILE "\n",'\i ',$env,'{'
	    
	    # Token in the ReadCommands command list that basic_lyx doesn't
	    #    know how to handle
	    } else {
		&copy_latex_known($eaten,$fileobject);
	    } # end big if

	    # Exit the switch
	    last TYPESW;
	}

	if (/^EndArgsToken$/) {
	    my $env = $eaten->[0]->[0]->[0];
	    print "$env" if $debug_on;

	    # Handle things that LyX translates as a "LatexCommand" inset
	    # \ref{blah} -> '\begin_inset LatexCommand \ref{blah}\n\n\end_inset'
	    if ($env =~ /^$LatexCommands$/o) {
	        print OUTFILE "\}\n\n\\end_inset\n";
	        last TYPESW; # skip to the end of the switch
	    }

	    if (exists $ReadCommands::ToLayout->{$env}) {
		&EndLayout($env);
		last TYPESW; # done translating

	    # Font characteristics
	    # Pop the current FontStatus stack for a given characteristic
	    #    and give the new command (e.g., \emph default)
	    } elsif (exists $FontTransTable{$env}) {
	        $command = $FontTransTable{$env};
	        ($dummy) = ($command =~ /(\S+)\s+\w+/);
		pop @{$CurrentFontStatus->{$dummy}};
		$command = "\n$dummy $CurrentFontStatus->{$dummy}->[-1]\n";
		print OUTFILE "$command";

	    # Footnotes and marginpars
	    } elsif (exists $FloatTransTable{$env}) {
	        print OUTFILE "\n\\end_float\n\n";

		# Reset the layout stack and font status table pointers to
		#    point to the global stack/table again, instead of the
		#    footnote-specific stack/table
		$CurrentFontStatus = \%FontStatus;
		$CurrentLayoutStack = \@LayoutStack;

		# We need to reissue any font commands (but not layouts)
		foreach $dummy (keys %$CurrentFontStatus) {
		    if ($CurrentFontStatus->{$dummy}->[-1] ne "default") {
			print OUTFILE $FontTransTable{$dummy};
		    }
		}

		# Same paragraph status as we had before the footnote started
		$IsNewParagraph = $OldINP; $MayBeDeeper = $OldMBD;
		$CurrentAlignment = $OldAlignment;

	    } elsif ($env =~ m/^$AccentTokens$/) {
		print OUTFILE "}\n";
	    
	    } # End if on $env

	    # Exit main switch
	    last TYPESW;
	} # end if EndArgsToken

	# Handle END of scope of local commands like \large
	if (/^EndLocal$/) {
	    $tok = $outstr; # here, outstr is the token, not the string
	    $dummy = "$tok->[0]"; #local cmd which started this, e.g.,\large
	    print $dummy if $debug_on;

	    if (exists $FontTransTable{$dummy}) {
		$command = $FontTransTable{$dummy};
		($dummy = $command) =~ s/\s*(\S*)\s+(\w+)\s*/$1/; #e.g., '\size'
		die "Font command error" if !exists $$CurrentFontStatus{$dummy};
		# TT::OF->check_presynthetic returns local commands FIFO!
		# So pop font stack, but warn if we pop the wrong thing
		warn " font confusion?" if
			   pop @{$CurrentFontStatus->{$dummy}} ne $2;
		print "\nCurrent $dummy Stack: @{$CurrentFontStatus->{$dummy}}"
		      if $debug_on;
		my $newfont = $CurrentFontStatus->{$dummy}->[-1];
		$command = "\n$dummy $newfont\n";
		print OUTFILE "$command";

	    } else {
	        warn "Unknown EndLocal token!\n";
	    }

	    last TYPESW;
	}

	# Handle '{'
	if (/^Begin::Group$/) {
	    #print "$outstr";
	    last TYPESW;
	}

	# Handle '{'
	if (/^End::Group$/) {
	    #print "$outstr";
	    last TYPESW;
	}

        # Handle \begin{foo}
        if (/^Begin::Group::Args$/) {
	    print "$outstr " if $debug_on;
	    $env = $eaten->environment;
	    print "$env" if $debug_on;
	    
	    # Any environment found in the layouts files
	    if (exists $ReadCommands::ToLayout->{$env}) {
	        &ConvertToLayout($env);

	    # Math environments
	    } elsif ($env =~ /^$MathEnvironments$/o) {
		&CheckForNewParagraph; # may be beginning of paragraph
		my $begin_text = $eaten->print;
		print "\nCopying math beginning with '$begin_text'\n"
		                                              if $debug_on;
		print OUTFILE "\n\\begin_inset Formula $begin_text ";
		$dummy = &copy_verbatim($fileobject, $eaten);
		print $dummy if $debug_on;
		print OUTFILE $dummy;

	    # Alignment environments
	    } elsif (exists $AlignEnvironments{$env}) {
		# Set it to the command which creates this alignment
	        $CurrentAlignment = $AlignEnvironments{$env};
		($dummy) = ($CurrentAlignment =~ /\S+\s+(\w+)/);
		print "\nNow $dummy-aligning text " if $debug_on;

		# alignment environments automatically start a new paragraph
		$IsNewParagraph = 1;

	    # Environments lyx translates to floats
	    } elsif (exists $FloatEnvTransTable{$env}) {
		$tok = $fileobject->eatOptionalArgument;
		if (defined ($dummy = $tok->[0])) {
		    print "\nIgnoring float placement '$dummy'" if $debug_on;
		}
	        $command = $FloatEnvTransTable{$env};

		# Open the footnote
		print OUTFILE "$command";

	    # table
	    } elsif ($env =~ /tabular$/) { # don't allow tabular*
		# Table must start a new paragraph
		$IsNewParagraph = 1; $MayBeDeeper = 1;
		# We want to print table stuff *after* a \layout Standard
		&CheckForNewParagraph;

		# Since table info needs to come *before* the table content,
		#    put a line in the output file saying that the *next*
		#    reLyX pass needs to put the table info there
		print OUTFILE "\n$RelyxTable::TableBeginString";

		# Read and ignore an optional argument [t] or [b]
		$tok = $fileobject->eatOptionalArgument;
		if (defined ($dummy = $tok->[0])) {
		    print "\nIgnoring positioning arg '$dummy'" if $debug_on;
		}

	        # Read the argument into a TT::Group
		#   (that group may contain groups, e.g. for {clp{10cm}}
		$tok = $fileobject->eatGroup;
		new RelyxTable::Table $tok;

	    # \begin document
	    } elsif ($env eq "document") {
	        print "\nStarting to translate actual document" if $debug_on;

	    # Special environments to copy as regular text (-r option).
	    # Do this by copying the \begin & \end command in TeX mode
	    } elsif (grep /^$env$/, @ReadCommands::regular_env) {
		print "\nCopying $env environment as regular text\n"
		                                              if $debug_on;
		&CheckForNewParagraph; # e.g., top of document
		$dummy = $eaten->print; # \begin{env}, ignore initial whitespace
		&print_tex_mode($dummy);

	    # otherwise, it's an unknown environment
	    # In that case, copy everything up to the \end{env}
	    #    Save stuff in global tex_mode_string so we can print it
	    # when we read & handle the \end{env}
	    } else {

		print "\nUnknown environment $env" if $debug_on;
		$tex_mode_string = $eaten->exact_print; # \begin{env}
		$tex_mode_string .= &copy_verbatim($fileobject,$eaten);
	    }

            last TYPESW;
        }
        
        # Handle \end{foo}
        if (/^End::Group::Args$/) {
	    print "$outstr " if $debug_on;
	    $env = $eaten->environment;
	    print "$env" if $debug_on;

	    # End of list or quote/verse environment
	    # OR special environment given with -t option
	    if (exists $ReadCommands::ToLayout->{$env}) {
	        &EndLayout($env);

	    # End of math environments
	    } elsif ($env =~ /^$MathEnvironments$/o) {
		print OUTFILE "\\end{$env}\n\\end_inset\n";
		print "\nDone copying math environment '$env'" if $debug_on;

	    } elsif (exists $AlignEnvironments{$env}) {
		# Back to block alignment
	        $CurrentAlignment = "";
		print "\nBack to block alignment" if $debug_on;
		# \end{} doesn't necessarily start new paragraph
		#     so we don't set $IsNewParagraph here

	    # Environments lyx translates to floats
	    } elsif (exists $FloatEnvTransTable{$env}) {
	        print OUTFILE "\n\\end_float\n\n";

	    # table
	    } elsif ($env =~ /tabular$/) { # don't allow tabular*
	        if ($thistable = &RelyxTable::in_table) {
		    $thistable->done_reading;
		    print OUTFILE $RelyxTable::TableEndString;
		} else {warn "found \\end{tabular} when not in table!"}

		# Anything after a table will be a new paragraph
		$IsNewParagraph = 1; $MayBeDeeper = 1;

	    } elsif ($env eq "document") {
	        print "\nDone with document!" if $debug_on;

	    # "regular" environment given with -r option
	    } elsif (grep /^$env$/, @ReadCommands::regular_env) {
		$dummy = $eaten->print; # \end{env}, ignore initial whitespace
		&print_tex_mode($dummy);

		# Next stuff will be new env.
		$IsNewParagraph = 1;

	    # End of unknown environments. We're already in TeX mode
	    } else {
		# Add \end{env} (including initial whitespace) to string
		$tex_mode_string .= $eaten->exact_print;
		# Now print it
		&print_tex_mode($tex_mode_string);
		print "Done copying unknown environment '$env'" if $debug_on;
	    }

            last TYPESW;

        }

	if (/^Text$/) {
	    last TYPESW if $outstr =~ /^\s+$/; #don't bother printing whitespace

	    # Only write '\layout Standard' once per paragraph
	    &CheckForNewParagraph;

	    # Print the LaTeX text to STDOUT
	    print $outstr if $debug_on;

	    # LyX runs together text separated by just \n, whereas LaTeX
	    #    considers \n just like a space.
	    # (But if there's "   \n   ", we still only need " \n".)
	    $outstr =~ s/[ \t]*\n[ \t]*/ \n/g;

	    # Many spaces (or tabs) are equivalent to one space in LaTeX
	    # (But LyX displays them on screen, so get rid of them)
	    $outstr =~ s/[ \t]+/ /g;

	    # protect spaces in an optional argument if necessary
	    if ($protect_spaces) {
		$dummy = $TextTokenTransTable{'~'};

		# This will not handle brackets in braces!
		if ($outstr =~ /\]/) {
		    # protect spaces only before the bracket
		    my $tempstr = $`;
		    my $tempstr2 = $';
		    $tempstr =~ s/[ \t]+/$dummy/g;
		    $outstr = $tempstr . $tempstr2; # don't print the bracket
		    $protect_spaces = 0; # Done with optional argument
		} else {
		    $outstr =~ s/[ \t]+/$dummy/g;
		}

	    }

	    # Translate any LaTeX text that requires special LyX handling
	    foreach $dummy (keys %TextTransTable) {
		$outstr =~ s/\Q$dummy\E/$TextTransTable{$dummy}/g;
	    }

	    # Actually print the text
	    print OUTFILE "$outstr";
	    last TYPESW;
	}

	# The default action - print the string.
	print "I don't know $outstr" if $debug_on;

    } # end for ($type)

    print "\n" if $debug_on;

} #end sub basic_lyx


#######################  VERBATIM COPYING SUBROUTINES  ########################
# Note: one or more of these subroutines may need to be moved elsewhere, if
#    they'll also be used by class-specific translating routines

sub copy_verbatim {
# This subroutine eats text verbatim until a certain text is reached
# The end text itself is not eaten; this is necessary so that
#    environments are properly nested (otherwise, TeX.pm complains)
# It returns a string containing the text
#
# Arg 0 is the Text::TeX::OpenFile file object, arg 1 is the beginning token
    my $fileobject = shift;
    my $begin_token = shift;
    my %endtokentbl = (  '\(' => '\)' , '\[' => '\]'  );

    my $type = ref($begin_token);
    $type =~ s/^Text::TeX:://o or die "unknown token type $type?!";

# Figure out beginning & end text of this token or environment
# Beginning text so we know if you have an environment nested within itself
# End text so we know when to finish copying OR when to 'pop' a level
#    if an environment is nested within itself
# Because the searches will generally be matching expressions with backslashes
#    and other meta-characters, we put \Q\E around (pieces of) the expressions
    my ($begin_text, $end_text);
    if ($type =~ /^Token$/) {  # \( or \[
	$begin_text = $begin_token->print; # e.g., '\('
        die "unknown begin_text" unless exists $endtokentbl{$begin_text};
	$end_text = "\Q$endtokentbl{$begin_text}\E";
	# actually, begin_text shouldn't be nec. since you can't nest math
	$begin_text = "\Q$begin_text\E"; # quote slashes, etc.

    } elsif (/^Begin::Group::Args$/) {    # \begin{foo}
	# \s* to allow, e.g., '\begin {foo}'
	$begin_text = $begin_token->print;
	$begin_text = "\Q$begin_text\E";
	$begin_text =~ s/begin/begin\\s*/;
        ($end_text = $begin_text) =~ s/begin/end/;

    } else {
        die "copy_verbatim called with unknown token type $type!";
    }
    #print "\nsub copy_verbatim going to copy until $end_text\n" if $debug_on;

# Actual copying
    my $textref; # reference to stuff we read in to print
    my $to_print = ""; #text to print
    # we're automatically "nested" once since we had the original \begin
    my $nest_count = 1;

    # (Eat and) Print out paragraphs until you find $end_text
    # paragraph returns "" if it's time to get a new paragraph -- if that
    #    happens, we want to continue, but we can't dereference $textref
    # Allow nesting of this environment!
    while (defined ($textref = $fileobject->paragraph)) {
	next unless $textref; # new paragraph; keep going

	# If we match begin or end text, eat everything up to it
	# Make sure to handle (nested) begin & end texts in order, so we can
	#    differentiate \begin \begin \end \end from \begin \end \begin \end
	if ($$textref =~ /$end_text/ && $` !~ /$begin_text/) {
	    # Note that $` will be from the last *successful* match,
	    #    namely the end_text match
	    --$nest_count;
	    $to_print .= $`; # print until the \end command
	    if ($nest_count) {
		$to_print .= $&; # print the end text too
		$$textref = $'; # leave the rest in the paragraph
	    } else {
		# Leave end text (and anything after it) for TeX.pm
		$$textref = $& . $';
		last; # done copying since there's no more nesting
	    }

	# If we match beginning text, we have a nested environment
	} elsif ($$textref =~ /$begin_text/ && $` !~ /$end_text/) {
	    $to_print .= $`; # print up to and
	    $to_print .= $&; # INCLUDING the begin text
	    $$textref = $'; # leave the rest in the paragraph
	    ++$nest_count;

	# If we didn't match begin OR end text, just eat the whole paragraph
	} else {
	    $to_print .= $$textref;
	    $$textref = "";
        } # end if $$textref
    } #end while
    die "eof without finding matching text $end_text" if (!defined $textref);

    # return the string
    #print "Exiting sub copy_verbatim\n" if $debug_on;
    return $to_print;
} # end copy_verbatim

# This subroutine copies and prints a latex token and its arguments if any.
# This sub is only needed if the command was not found in the syntax file
# Use exact_print to preserve, e.g., whitespace after macros
sub copy_latex_unknown {
    my $eaten = shift;
    my $fileobject = shift;
    my $outstr = $eaten->exact_print;
    my ($dummy, $tok, $count);

# Copy the actual word. Copy while you've still got
#     arguments. Assume all args must be in the same paragraph
#     (There could be new paragraphs *within* args)
    # We can't use copy_verbatim (unless we make it smarter) because
    #     it would choke on nested braces
    # If unknown stuff is just '\foo\n\n' lookAheadToken will return ''
    print "\nUnknown token: '",$eaten->print,"': Copying in TeX mode\n"
                                                         if $debug_on;
    my $dum2;
    while (($dum2 = $fileobject->lookAheadToken) &&
	   ($dum2 =~ /^[[{]$/)) {
	if ($dum2 eq '[') { #copy optional argument - assume it's simple
	    $tok = $fileobject->eatOptionalArgument;
	    $outstr .= $tok->exact_print; # also print brackets & whitespace
	} else {
	    $count = 0;
	    EAT: { #copied from eatBalanced, but allow paragraphs
	        die unless defined ($tok = $fileobject->eatMultiToken);
		$outstr.="\n",redo EAT if ref($tok) eq "Text::TeX::Paragraph";
		$dummy = $tok->exact_print;
		$outstr .= $dummy;
		$count++ if $dummy eq '{'; # add a layer of nesting
		$count-- if $dummy eq '}'; # end one layer of nesting
		redo EAT if $count; #don't dump out until all done nesting
	    } #end EAT block
	} # end if $dummy = [{

    } #end while
    # Add {} after macro if it's followed by '}'. Otherwise, {\foo}bar
    #     will yield \foobar in the temp. file made by LyX
    $outstr.="{}" if $outstr=~/\\[a-zA-Z]+$/ && $dum2 eq '}';

    # Print it out in TeX mode
    &print_tex_mode($outstr);

    print "Done copying unknown token" if $debug_on;
} # end sub copy_latex_unknown

# Copy a latex argument whose syntax we know
#    Once this gets initialized (with a BegArgsToken), it will be called with
# lots of tokens, as the parser returns them. When the EndArgsToken which
# corresponds to the BegArgsToken appears, tell the program it doesn't need
# to be calling this sub any more.
sub copy_latex_known {
    my ($eaten, $fileobject) = (shift,shift);
    my $type = ref($eaten);
    $type =~ s/^Text::TeX::// or die "unknown token?!";

# Initialize tex copying if we're not already copying in tex mode
    if (!$in_tex_mode) {

	print "\nCopying untranslatable token '",$eaten->print,
	                              "' in TeX mode\n" if $debug_on;
	# Token that started verbatim copying
	# (== token itself if TT::Token, base token for TT:BegArgsToken)
	$tex_start = $eaten->base_token;
	# initialize the string of stuff we're copying
	$tex_mode_string = "";
	# BegArgsToken: we need to copy all the arguments in tex mode
	# Otherwise (regular Token) just copy this token in tex mode,
	#    possibly with optional arguments
	$in_tex_mode = ($type eq "BegArgsToken");
    } # Start tex copying?

# We'll want to print the token
    $tex_mode_string .= $eaten->exact_print;

# Now do some special stuff only if it's the token that started the
# tex mode copying (or an ArgToken or EndArgsToken of that token)
    if ($type =~ /Token/ && ($eaten->base_token == $tex_start)) {

	# If there could be optional arguments next, copy them
	# If next token is not '[' (or it's end of par), stop looking for args
	my $curr_args = $eaten->next_args($fileobject);
	while ($curr_args =~ s/^o// && $fileobject->lookAheadToken eq '[') {
	    my $opt = $fileobject->eatOptionalArgument;
	    $tex_mode_string .= $opt->exact_print;
	}

	# We're done if we get the corresponding EndArgsToken
	$in_tex_mode = 0 if $type =~ /EndArgsToken$/;
    } # end doing special stuff for token that started tex mode

# Finish copying in tex mode? This can happen either because we got
# an EndArgsToken OR because we were just copying a regular Token and
# never set in_tex_mode in the first place
    if (!$in_tex_mode) {

	# Print out the string, start new paragraph if necessary, etc.
	&print_tex_mode($tex_mode_string);

	print "Done copying known token and its arguments:\n" if $debug_on;
    } # end tex copying?
} # end sub copy_latex_known

# Print a string in LyX "TeX mode"
#    The goal of this subroutine is to create a block of LyX which will be
# translated exactly back to the original when LyX creates its temporary LaTeX
# file prior to creating a dvi file.
#    Don't print \n\n at the end of the string... instead just set the
sub print_tex_mode {
    my $outstr = shift;
    my $str_ends_par = ($outstr =~ s/\n{2,}$//);

    if ($debug_on) {
       print "\n" unless ($outstr =~ /^\n/);
       print $outstr; print "\n" unless ($outstr =~ /\n$/)
    }

    # Check whether string came right at the beginning of a new paragraph
    &CheckForNewParagraph;

    # Get into TeX mode
    print OUTFILE "\n\\latex latex\n";

    # Do TeX mode translation;
    $outstr =~ s/\\/\n\\backslash /g;
    # don't translate \n in '\n\backslash' that we just made!
    $outstr =~ s/\n(?!\\backslash)/\n\\newline\n/g;

    # Print the actual token + arguments if any
    print OUTFILE $outstr;

    # Get OUT of LaTeX mode (and end nesting if nec.)
    print OUTFILE "\n\\latex default\n";
    $IsNewParagraph = $str_ends_par;

    return;
} # end sub print_tex_string

#########################  TRANSLATOR SUBROUTINES  ########################
sub CheckForNewParagraph {
# This subroutine makes sure we only write \layout command once per paragraph
#    It's mostly necessary cuz 'Standard' layout is the default in LaTeX;
#    there is no command which officially starts a standard environment
# If we're in a table, new layouts aren't allowed, so just return
# If $IsNewParagraph is 0, it does nothing
# If $INP==1, It starts a new paragraph
# If $CurrentAlignment is set, it prints the alignment command for this par.
# If $MayBeDeeper==1 and we're currently within a list environment,
#    it starts a "deeper" Standard paragraph
    my $dummy; 
    my $layout = $$CurrentLayoutStack[-1];

    return if &RelyxTable::in_table;

    # Handle standard text within a list environment specially
    if ($MayBeDeeper) {
        if ($layout =~ /^$ListLayouts$/o) {
	    push (@$CurrentLayoutStack, "Standard");
	    print "\nCurrent Layout Stack: @$CurrentLayoutStack\n" if $debug_on;
	    print OUTFILE "\n\\begin_deeper\n";
	    $layout = "Standard";
	}
	$MayBeDeeper = 0; # Don't test again until new paragraph
    }

    # Now write the new layout and alignmnet & font commands if we need to
    if ($IsNewParagraph) {
	print OUTFILE "\n\\layout $layout\n\n";
	print OUTFILE $CurrentAlignment if $CurrentAlignment;

	# Now that we've written the command, it's no longer a new paragraph
	$IsNewParagraph = 0;

	# When you write a new paragraph, reprint any font commands
	foreach $dummy (keys %$CurrentFontStatus) {
	    my $currf = $CurrentFontStatus->{$dummy}->[-1];
	    if ($currf ne "default") {
		print OUTFILE "\n$dummy $currf\n";
	    }
	}
    } # end if $INP
} # end sub CheckForNewParagraph

sub ConvertToLayout {
# This subroutine begins a new layout, pushing onto the layout stack, nesting
# if necessary. It doesn't actually write the \layout command -- that's
# done by CheckForNewParagraph.
#    The subroutine assumes that it's being passed an environment name or macro
# which is known and which creates a known layout
#    It uses the ToLayout hash (created by the ReadCommands module) which
# gives the LyX layout for a given LaTeX command or environment
# Arg0 is the environment or macro
    my $name = shift;

    my $layoutref = $ReadCommands::ToLayout->{$name};
    my $layout = $layoutref->{'layout'};
    my $dummy = ($name =~ /^\\/ ? "macro" : "environment");
    print "\nChanging $dummy $name to layout $layout" if $debug_on;

    # Nest if the layout stack has more than just "Standard" in it
    if ($#{$CurrentLayoutStack} > 0) {
	# Die here for sections & things that can't be nested!
	print " Nesting!" if $debug_on;
	print OUTFILE "\n\\begin_deeper\n";
    }

    # Put the new layout onto the layout stack
    push @$CurrentLayoutStack, $layout;
    print "\nCurrent Layout Stack: @$CurrentLayoutStack"
	  if $debug_on;

    # Upcoming text will be new paragraph, needing a new layout cmd
    $IsNewParagraph = 1;

    # Test for nested "Standard" paragraph in upcoming text?
    # Some environments can nest. Sections & Title commands can't
    $MayBeDeeper = $layoutref->{"nestable"};

} # end sub ConvertToLayout

sub EndLayout {
# This subroutine ends a layout, popping the layout stack, etc.
#    The subroutine assumes that it's being passed an environment name or macro
# which is known and which creates a known layout
#    It uses the ToLayout hash (created by the ReadCommands module) which
# gives the LyX layout for a given LaTeX command or environment
# Arg0 is the environment or macro
    my $name = shift;

    my $layoutref = $ReadCommands::ToLayout->{$name};
    my $layout = $layoutref->{'layout'};
    my $dummy = ($name =~ /^\\/ ? "macro" : "environment");
    print "\nEnding $dummy $name (layout $layout)" if $debug_on;

    my $mylayout = pop (@$CurrentLayoutStack);

    # If a standard paragraph was the last thing in a list, then
    #     we need to end_deeper and then pop the actual list layout
    # This should only happen if the Standard layout was nested
    #    in a nestable layout. We don't check.
    if ($mylayout eq "Standard") {
	print OUTFILE "\n\\end_deeper\n";
	print " End Standard Nesting!" if $debug_on;
	$mylayout = pop (@$CurrentLayoutStack);
    }

    # The layout we popped off the stack had better match the
    #    environment (or macro) we're ending!
    if ($mylayout ne $layout) { die "Problem with Layout Stack!\n"};
    print "\nCurrent Layout Stack: @$CurrentLayoutStack" if $debug_on;

    # If we're finishing a nested layout, we need to end_deeper
    # This should only happen if the layout was nested
    #    in a nestable layout. We don't check.
    # Note that in we're nested in a list environment and the
    #     NEXT paragraph is Standard, then we'll have an extra
    #     \end_deeper\n\begin_deeper in the LyX file. It's sloppy
    #     but it works, and LyX will get rid of it when it
    #     resaves the file.
    if ($#{$CurrentLayoutStack} > 0) {
	print " End Nesting!" if $debug_on;
	print OUTFILE "\n\\end_deeper\n";
    }

    # Upcoming text will be new paragraph, needing a new layout cmd
    $IsNewParagraph = 1;

    # Test for nested "Standard" paragraph in upcoming text?
    # Some environments can nest. Sections & Title commands can't
    $MayBeDeeper = $layoutref->{"nestable"};
} # end sub EndLayout

1; # return true to calling subroutine

