# This file is part of reLyX, copyright (c) 1998 Amir Karger
# You are free to use and modify this code under the terms of
# the GNU General Public Licence version 2 or later.

package ReadCommands;
# Read a file containing LaTeX commands and their syntax
# Also, read a file containing LyX layouts and their LaTeX equivalents

use strict; 

# Variables needed by other modules
# ToLayout is a ref to a hash which contains LyX layouts & LaTeX equivalents
# regular_env is a list of environments that have "reLyXable" LaTeX in them
use vars qw($ToLayout @regular_env);

#    CommandHash is the hash of LaTeX commands and their syntaxes which
# this package builds.
my %CommandHash;

# Name of the environment containing names of regular environments :)
my $regenv_name = "reLyXre";
# This variable is set when reading regular environments from syntax files
my $in_regular_env;

#########################  READ COMMAND SYNTAX  ################################
sub read_syntax_files {
# This subroutine calls the TeX parser & translator to read LaTeX syntax file(s)
#    It sets the list of "regular" environments, environments which aren't known
# by reLyX, but which contain text that reLyX can translate.
#
# @_ contains the syntax file(s) to read

    my @syntaxfiles = @_;

# Before anything else, set the package-wide variables based on the
#    user-given flags
    # opt_d is set to 1 if '-d' option given, else (probably) undefined
    my $debug_on = (defined($main::opt_d) && $main::opt_d);

    # opt_r is a group of environments to copy like regular text
    # If opt_r wasn't given, then there are no such special environments
    @regular_env = (defined $main::opt_r ? (split(/,/, $main::opt_r)) : () );

    # The only important commands to pass are \begin and \end, so that the
    # parser realizes they start/end environments, as opposed to being
    # regular old tokens.
    my %MyTokens = ( '{' => $Text::TeX::Tokens{'{'},
                     '}' => $Text::TeX::Tokens{'}'},
                     '\begin' => $Text::TeX::Tokens{'\begin'},
                     '\end' => $Text::TeX::Tokens{'\end'},
    );
    
    my $InFileName;
    foreach $InFileName (@syntaxfiles) {
	die "could not find syntax file $InFileName" unless -e $InFileName;
	warn "Reading LaTeX commands from $InFileName... \n";

	# Open the file to turn into LyX.
	my $infile = new Text::TeX::OpenFile $InFileName,
	    'defaultact' => \&read_commands,
	    'tokens' => \%MyTokens;

	# When we start (each file), we're not reading regular environments yet
	$in_regular_env = 0;

	# Process the file
	$infile->process;
    }
    print "Regular environments: @regular_env\n" if $debug_on;

    warn "Done reading commands\n";
    return;
} # end subroutine call_parser

sub read_commands {
# This subroutine is called by Text::TeX::process
# Arg0 is the token we just ate
# Arg1 is the file object we're reading from
#
#    We create a hash, where each command is a key. The value is just a string
# of zero or more 'o' and 'r' characters. Each 'o' stands for an optional
# argument, 'r' stands for a required argument
#    In addition, the $regenv_name environment contains
# regular environments, like those input with the -r option.
#    Note that if a command is found more than once, then it wil be overwritten.
# This is a feature. This way, a user-defined syntax file can overwrite the
# argument list found in the default syntax file.
    my ($token,$fileobject) = (shift,shift);

    my $type = ref($token);
    $type =~ s/^Text::TeX::// or die "unknown token type $type from Text::TeX";

    # Because there's no token list, ALL tokens will be
    #    Paragraph, Text, or Token
    SWITCH: for ($type) {
       # Handle blank lines.
        if (/Paragraph/) {
	    # don't do anything
	    last SWITCH;

        } elsif (/^Token/) {
	    my ($dum2);
	    my $args = "";
	    while (($dum2 = $fileobject->lookAheadToken) &&
		   ($dum2 =~ /^[[{]$/)) {
		if ($dum2 eq '[') { #eat optional argument - assume it's simple
		    $fileobject->eatOptionalArgument;
		    $args .= "o";
		} else {
		    my $tok = $fileobject->eatBalanced or die "bad group";
		    #print "group is ",$tok->print,"\n";
		    $args .= "r";
		} # end if $dummy = [{
	    }
	    $CommandHash{$token->print} = $args;

	    # When we exit the loop, we got the end of a Paragraph, or we
	    # found the text at the end of the command (which includes \n)
	    # so we should just continue
	    last SWITCH;

        } elsif (/^Begin::Group::Args/) {
	    my $env = $token->environment;
	    die "Unknown environment" unless $env eq $regenv_name;
	    $in_regular_env = 1;
	    last SWITCH;
        } elsif (/^End::Group::Args/) {
	    my $env = $token->environment;
	    die "Unknown environment" unless $env eq $regenv_name;
	    $in_regular_env = 0;
	    last SWITCH;

        } elsif (/^Text/) {
	    # don't do anything unless we're reading environments
	    if ($in_regular_env) {
		my @new_envs = (split(/\s+/, $token->print));
		@new_envs = grep ($_, @new_envs); # remove empty elements
		push @regular_env,@new_envs;
	    }
	    last SWITCH;
        } else {
	    die "unexpected token type $type";
	}

    } # end SWITCH

} # end sub read_commands

sub Merge {
#    This sub creates a token list (which could be used to call a Text::TeX
# parser) from %CommandHash, and merges it with the input token list
#    If a command takes any required arguments, it will be a report_args,
# but if it just takes an optional argument, it can stay a regular old token.
# In either case, we insert a new field, "relyx_args", into the token list,
# which is the expected order of arguments for that command. Even if there
# are no args, we insert an empty relyx_args, so that we can differentiate
# between a truly unknown token and a known token which takes no args.
#
# We don't allow a new command to override a command that already exists in
# OldHash, i.e., one that was defined explicitly in the calling sub.
#
# Arg0 is a (reference to an) existing token list
    my $OldHashRef = shift;

    foreach (keys %CommandHash) {
	my $val = $CommandHash{$_};
	my ($entry, $count, @foo);
        if (!exists $OldHashRef->{$_}) {
	    if ($count = scalar(@foo = ($val =~ /r/g))) {
		# TeX.pm will make this a TT::BegArgsToken and $count-2
		#    TT::ArgTokens, followed by a TT::EndArgsToken
		$OldHashRef->{$_} = {"Type" => 'report_args',
		                     "count" => $count,
				     "relyx_args" => $val};
	    } else { # only non-required args
		# Make it a regular TT::Token, but write relyx_args
		#    (even if $val is "")
		$OldHashRef->{$_} = {"relyx_args" => $val};
	    }
	}
    } # end foreach

} # end sub Merge

############################  READ LAYOUTS  ####################################
sub read_layout_files {
# This subroutine reads a textclass-specific layout file and all files
#     included in that file.
# It sets up the layout hash table. For each environment, it describes which
#     layout that environment refers to. It does the same for macros which
#     begin LyX layouts (e.g., \section)
# 
# Arg0 is the name of the documentclass
    use FileHandle;
    use File::Basename;
    my $doc_class = shift;
    my @filestack;
    my $fh;
    my $line;
    # ToLayout{latexname} stores layout; so ReversHash{layout} = latexname
    my %ReverseHash;
    my $debug_on = (defined($main::opt_d) && $main::opt_d);

    # look for layout file in $HOME/.lyx first, then system layouts directory
    my $searchname = "$doc_class.layout";
    my @searchdirs = ();
    my $personal_layout = "$main::dot_lyxdir/layouts";
    push(@searchdirs,$personal_layout) if -e $personal_layout;
    my $system_layout = "$main::lyxdir/layouts";
    # I guess this won't exist if running reLyX without installing...
    # Of course, in that case, this will probably break
    push(@searchdirs,$system_layout) if -e $system_layout;
    my @foundfiles = grep(-e "$_/$searchname", @searchdirs) or
          die "Cannot find layout file $searchname in dir(s) @searchdirs";
    my $LayoutFileName = "$foundfiles[0]/$searchname"; # take first one we found

    $fh = new FileHandle;
    $fh->open ("<$LayoutFileName");
    warn "Reading layout file $LayoutFileName\n";
    push @filestack, $fh;

    # Read the layout file!
    my ($lyxname, $latexname, $latextype) = ("","","");
    my $fname;
    while() {
	# Read a line. If eof, pop the filestack to return to the file
	#    that included this file *or* finish if the stack's empty
        unless (defined ($line = <$fh>)) {
	    $fh->close;
	    pop @filestack;
	    last unless ($#filestack+1); # finish when stack is empty
	    $fh = $filestack[-1];
	    next; # read another line from the "calling" file
	}

	# This set of ifs deals with lines outside a style definition
        if ($line =~ /^Style\s*(\S+)/) { # start a style definition
	    $lyxname = $1;
	    $latexname = ""; # make sure these variables are unset
	    $latextype = "";
	} elsif ($line =~ /^Input\s*(\S+)/) { #include a file
	    $searchname = $1;
	    @foundfiles = grep(-e "$_/$searchname", @searchdirs) or
	      die "Cannot find layout file $searchname in dir(s) @searchdirs";
	    $fname = "$foundfiles[0]/$searchname"; # take first one we found
	    $fh = new FileHandle;
	    push @filestack, $fh;
	    $fh->open("<$fname");
	    print "Reading included layout file $fname\n" if $debug_on;
	}

	next unless $lyxname; # not w/in style definition

	# This set of ifs deals with lines within a Style definition
	if ($line =~ /^\s*LatexName\s*(\S+)/) {
	    $latexname = $1;
	    next;
	} elsif ($line =~ /^\s*LatexType\s*(\S+)/) {
	    $latextype = $1;
	    next;
	} elsif ($line =~ /^\s*CopyStyle\s*(\S+)/) { # copy an existing style
	    # "if" is necessary in case someone tries "CopyStyle Standard"
	    if (exists $ReverseHash{$1}) {
		my $layref = $ReverseHash{$1};
		$latexname = $layref->{"name"};
		$latextype = $layref->{"type"};
	    }

	# When you get to the end of a definition, create hash table entries
	#    (if you've found the right information)
	} elsif ($line =~ /^End\s*$/) {

	    if ($latextype and $latexname) {
		# Create reverse hash entry (needed for CopyStyle)
		# Do it before making modifications to $latexname, e.g.
		$ReverseHash{$lyxname} = {"name" => $latexname,
		                          "type" => $latextype};

		my ($nest, $skip) = (0,0);
		for ($latextype) { # make $_=$latextype
		    if (/^Command/) {
			# Macros need a '\' before them. Environments don't
			$latexname = '\\' . $latexname;

		    } elsif (/^Environment/) {
			$nest = 1;
		    } elsif (/Item_Environment/i || /List_Environment/) {
			$nest = 1;

		    # layout Standard has LatexType Paragraph. It shouldn't
		    #    have any hash entry at all
		    } elsif (/^Paragraph$/) {
		        $skip = 1;
		    } else {
			warn "unknown LatexType $latextype" . 
			     "for $latexname (layout $lyxname)!\n";
		    }
		} #end for

		# Create the hash entry
		unless ($skip) {
		    $ToLayout->{$latexname} = {"layout" => $lyxname,
						    "nestable" => $nest};
		}

		# Now that we've finished the style, unset $lyxname so that
		#     we'll skip lines until the next style definition
		$lyxname = "";
	    } # end if ($latextype and $latexname)
	} # end if on $line
	    
    } #end while

## Print every known layout
#    print "LatexName\tLayout\tNestable?\n";
#    foreach (sort keys %$ToLayout) {
#        print "$_\t$ToLayout->{$_}{'layout'}\t" .
#    	      "$ToLayout->{$_}{'nestable'}\n";
#    };
    warn "Done reading layout files\n";
    return;
} # end sub read_layout_files



1; # return TRUE to calling routine
