#!/usr/bin/perl -w
#
# Main code for reLyX - the LaTeX to LyX translator
#
# reLyX is copyright (c) 1998 Amir Karger
# You are free to use and modify this code under the terms of
# the GNU General Public Licence version 2 or later.
#
# This code usually gets called by the reLyX wrapper executable
#
# $Id: reLyXmain.pl,v 1.4 1998/05/25 13:08:18 lasgoutt Exp $
#

require 5.002; # Perl 5.001 doesn't work. Perl 4 REALLY doesn't work.

# Standard Perl Library modules
use strict; # must define all variables, no barewords, no hard references

# Variables from other files, like Getopt::Std, OR vars that need to be global
use vars qw($opt_c $opt_d $opt_f $opt_r $opt_s $dot_lyxdir);
use Getopt::Std; # read in command-line options
use File::Basename; # &basename and &dirname

# Modules containing subroutines for reLyX
# Note that @INC must include the directory that these modules are in
use Text::TeX; # TeX parser package
use ReadCommands; # package to read LaTeX commands' syntax from a file
use MakePreamble; # package to split off LaTeX preamble & translate it to LyX
use CleanTeX; # package to clean TeX file for Lyxifying
use BasicLyX; # package to translate clean TeX to Basic LyX
use LastLyX; # package to print out LyX file once all translation is done


# Hack to allow running reLyX without the wrapper
if (!defined($lyxdir)) {$lyxdir = "/usr/local/share/lyx"}
if (!defined($lyxname)) {$lyxname = "lyx"}
# there's a use vars in the wrapper, so we only need these if running w/out it
use vars qw($lyxdir $lyxname);
 
# variables that a user might want to change
my @suffixlist = '\.(ltx|latex|tex)'; # allowed suffixes for TeX file
my $LyXFormat = "2.15"; #What to print in \lyxformat command in .lyx file
my $syntaxname = "syntax.default"; # name of the default syntax file
$dot_lyxdir = $ENV{'HOME'} . "/.$lyxname"; # personal .lyx directory

##############################################################################
# MAIN
#

# Important variables!
my ($InFileName, $OutFileName); # current in/output file names
my $CommandsFileName; # File with LaTeX command syntax
my $PreambleName; # name of the file we store the LaTeX preamble in
my $PathBase; # InFileName without .tex ending. For building output filenames
my @deletelist; # files to delete when we're done (if no '-d' flag)
my $doc_class; # LaTeX documentclass
my $LyXPreamble; # String in which we hold the *entire* LyX Preamble

#get arguments to reLyX, including TeX file name
($InFileName, $PathBase) = &getargs;

# Read file(s) containing LaTeX commands and their syntax
# Read personal syntax.default, or system-wide if there isn't a personal one
# Then read other syntax files, given by the -s option
my $default_file = "$dot_lyxdir/reLyX/$syntaxname";
if (! -e $default_file) {
    $default_file = "$lyxdir/reLyX/$syntaxname";
    die "cannot find default syntax file $default_file" unless -e $default_file;
}
my @syntaxfiles = ($default_file);
push (@syntaxfiles, (split(/,/,$opt_s))) if defined $opt_s;
&ReadCommands::read_syntax_files(@syntaxfiles);

# Split the preamble off of the rest of the file
# Copy until \begin{document} into foo.preamble
# Copy rest into foo.relyx1
$PreambleName = $PathBase . ".preamble";
push @deletelist, $PreambleName; # delete foo.preamble when done
my $filenum = 1; #numbering for temporary files
$OutFileName = $PathBase . ".relyx" . $filenum;
push @deletelist, $OutFileName;
&MakePreamble::split_preamble($InFileName, $PreambleName, $OutFileName);

# Now read and translate the LaTeX preamble into LyX
# Return the document's textclass so we know which layout file(s) to read
# Also return the LyX preamble in a string, for later
($doc_class,$LyXPreamble) =
         &MakePreamble::translate_preamble($PreambleName, $LyXFormat);

# Clean the TeX file (not including its preamble)
$InFileName = $OutFileName;
$OutFileName = $PathBase . ".relyx" . ++$filenum; # foo.relyx1 -> foo.relyx2
push @deletelist, $OutFileName;
&CleanTeX::call_parser($InFileName, $OutFileName);

# Read file(s) containing the valid LyX layouts for this documentclass and their
# LaTeX command/environment equivalents
&ReadCommands::read_layout_files($doc_class);

# Now convert basic constructs in the cleaned TeX file to LyX constructs
$InFileName = $OutFileName;
$OutFileName = $PathBase . ".relyx" . ++$filenum; # foo.relyx2 -> foo.relyx3
push @deletelist, $OutFileName;
&BasicLyX::call_parser($InFileName, $OutFileName);

# Someday, there will be calls to class-specific translation routines here
# Or will there?...

# Finally, print out the actual LyX file including the preamble
$InFileName = $OutFileName;
$OutFileName = $PathBase . ".lyx";
&LastLyX::last_lyx($InFileName, $OutFileName, $LyXPreamble);

unless ($opt_d) {
    warn "Deleting @deletelist\n";
    unlink @deletelist;
}

warn "Finished successfully!\n";

##################### SUBROUTINES  ###########################################
# Gets the script's arguments.
# Returns the input file name, as well as the 'pathbase', which
#    is the input file name (including path) without the '.tex' ending.
#    It will be used for building temporary and output file names
sub getargs {
    # set $opt_f if '-f' on command line, "r:" means -r takes an argument
    #   in which case $opt_r is set to the argument
    getopts('c:dfr:s:') or die "Illegal option! man reLyX for details\n";
    my $InFileName = $ARGV[0];
    die "No filename specified.\n" unless defined $InFileName;

    # Get file names, set up for making different temporary files
    # fileparse_set_fstype("MSDOS") for DOS support!
    my ($basename, $path, $suffix) = fileparse($InFileName, @suffixlist);
    $path .= '/' unless $path =~ /\/$/; # fix BUG in perl5.002 fileparse!

    # Try adding .tex to filename if you can't find the file the user input
    unless (-e $InFileName) {
        if (! $suffix) { # didn't have a valid suffix. Try adding one
	    if (-e "$InFileName.tex") {
		$suffix = ".tex";
	        $InFileName .= $suffix;
	        print "Using input file $InFileName\n" if $opt_d;
	    } else {
	 	die "Cannot find input file $InFileName or $InFileName.tex\n"
	    }
	} else { # it had a valid suffix, but the file doesn't exist
	    die "Cannot find input file $InFileName\n"
	}
    }
    my $PathBase = $path . $basename; # all but suffix

    # Check for files that already exist
    my $pname = $PathBase . ".preamble";
    my $lname = $PathBase . ".lyx";
    if ($opt_f) {
	warn "Will overwrite Preamble file $pname\n" if -e $pname;
	warn "Will overwrite LyX file $lname\n" if -e $lname;
    } else {
	die "Preamble file $pname already exists\n" if -e $pname;
	die "LyX file $lname already exists\n" if -e $lname;
    }

    return ($InFileName, $PathBase);
}


1; # return "true" to the wrapper
